'-------------------------------------------------------------------------------
' FBServer.bas - Demo server program
'
' Brief description:
' The Server "listens" on a non-blocking (master) socket, so that WndProc Msgs can be serviced.
' A request for connection from a remote computer generates a WndProc Msg with an FD_ACCEPT event.
' A ConnPending flag is set. When this flag is acted upon (sktAccept() fn is called), a NEW
' socket (Client) is created (by the system) with IDENTICAL settings to the master socket.
'
' This NEW Client socket is non-blocking, and is reconfigured to ONLY have an FD_CLOSE event active.
' The Client socket descriptor (handle) is used in all further comms with the remote computer.
'
' The master socket remains in the "listening" state, waiting for new connections.
' 
' FreeBasic code generated by RapidFRM v1.61
'-------------------------------------------------------------------------------
#INCLUDE "windows.bi"
#INCLUDE "win/commctrl.bi"
InitCommonControls()

#DEFINE __QSOCKET
#INCLUDE ONCE "win/winsock.bi"
#INCLUDE "win/RQInclude.bi"

'
' My global socket stuff
'
#DEFINE SERVER_PORT 5000
#DEFINE MAX_CLIENTS 64

DIM SHARED AS integer ListenSocket                  ' Master listening socket (handle)
DIM SHARED AS integer Clients(1 TO MAX_CLIENTS)     ' Remote connection socket handles
DIM SHARED AS integer NumClients=0

'-------------------------------------------------------------------------------
' Global data from Form designer sheet
'-------------------------------------------------------------------------------
DIM SHARED AS QFORM Form1
DIM SHARED AS QFORM PTR lpForm1 = @Form1
DIM SHARED AS QLISTBOX ListBox1
DIM SHARED AS QLISTBOX PTR lpListBox1 = @ListBox1
DIM SHARED AS QBUTTON Butn2
DIM SHARED AS QBUTTON PTR lpButn2 = @Butn2
DIM SHARED AS QBUTTON Butn3
DIM SHARED AS QBUTTON PTR lpButn3 = @Butn3
DIM SHARED AS QTIMER Timer1
DIM SHARED AS QTIMER PTR lpTimer1 = @Timer1
DIM SHARED AS QSOCKET Socket1
DIM SHARED AS QSOCKET PTR lpSocket1 = @Socket1

'-------------------------------------------------------------------------------
' Forward declarations for component events 
'-------------------------------------------------------------------------------
DECLARE SUB StartServer(Sender AS QBUTTON)
DECLARE SUB DoExit(Sender AS QBUTTON)
DECLARE SUB DoTimer(Sender AS QTIMER)

'-------------------------------------------------------------------------------
' Configure components
'-------------------------------------------------------------------------------
SUB ConfigureComponents

    lpForm1->ObjIndex = 0
    lpForm1->Width = 400
    lpForm1->Height = 300
    lpForm1->Caption = "FB Server"

    lpListBox1->ObjIndex = 1
    lpListBox1->Left = 10
    lpListBox1->Top = 10
    lpListBox1->Width = 370
    lpListBox1->Height = 180
    lpListBox1->Parent = Form1.Handle

    lpButn2->ObjIndex = 2
    lpButn2->Left = 60
    lpButn2->Top = 220
    lpButn2->Caption = "Start"
    lpButn2->Parent = Form1.Handle
    lpButn2->OnClick = CAST(ANY PTR, @StartServer)

    lpButn3->ObjIndex = 3
    lpButn3->Left = 240
    lpButn3->Top = 220
    lpButn3->Caption = "EXIT"
    lpButn3->Parent = Form1.Handle
    lpButn3->OnClick = CAST(ANY PTR, @DoExit)

    lpTimer1->ObjIndex = 4
    lpTimer1->Interval = 250
    lpTimer1->OnTimer = CAST(ANY PTR, @DoTimer)

    lpSocket1->ObjIndex = 5
    lpSocket1->Parent = Form1.Handle
    lpSocket1->Protocol = IPPROTO_TCP
END SUB

'-------------------------------------------------------------------------------
' Create components
'-------------------------------------------------------------------------------
SUB AddChildObjects
DIM AS UINT StdStyle, ExtStyle
DIM AS string ObjCaption

    lpListBox1->PreSetConfig()
    ObjCaption=""
    lpListBox1->Parent=lpForm1->Handle
    StdStyle=lpListBox1->objStyle
    IF lpListBox1->Visible THEN StdStyle += WS_VISIBLE
    IF lpListBox1->BorderStyle=bsNone THEN ExtStyle=0 ELSE ExtStyle=WS_EX_CLIENTEDGE

    lpListBox1->Handle=CreateWindowEx(ExtStyle,lpListBox1->objClass, ObjCaption, WS_CHILD OR StdStyle, _
                    lpListBox1->Left, lpListBox1->Top, lpListBox1->Width, lpListBox1->Height, _
                    lpForm1->Handle, CAST(HMENU, lpListBox1->objIndex), 0, 0)

    lpListBox1->PostConfig()
    EnableWindow(lpListBox1->Handle, lpListBox1->Enabled)

    lpButn2->PreSetConfig()
    ObjCaption=lpButn2->Caption
    lpButn2->Parent=lpForm1->Handle
    StdStyle=lpButn2->objStyle
    IF lpButn2->Visible THEN StdStyle += WS_VISIBLE
    ExtStyle=0

    lpButn2->Handle=CreateWindowEx(ExtStyle,lpButn2->objClass, ObjCaption, WS_CHILD OR StdStyle, _
                    lpButn2->Left, lpButn2->Top, lpButn2->Width, lpButn2->Height, _
                    lpForm1->Handle, CAST(HMENU, lpButn2->objIndex), 0, 0)

    lpButn2->PostConfig()
    EnableWindow(lpButn2->Handle, lpButn2->Enabled)

    lpButn3->PreSetConfig()
    ObjCaption=lpButn3->Caption
    lpButn3->Parent=lpForm1->Handle
    StdStyle=lpButn3->objStyle
    IF lpButn3->Visible THEN StdStyle += WS_VISIBLE
    ExtStyle=0

    lpButn3->Handle=CreateWindowEx(ExtStyle,lpButn3->objClass, ObjCaption, WS_CHILD OR StdStyle, _
                    lpButn3->Left, lpButn3->Top, lpButn3->Width, lpButn3->Height, _
                    lpForm1->Handle, CAST(HMENU, lpButn3->objIndex), 0, 0)

    lpButn3->PostConfig()
    EnableWindow(lpButn3->Handle, lpButn3->Enabled)

    lpTimer1->Parent=lpForm1->Handle
    SetTimer(lpForm1->Handle, lpTimer1->ObjIndex, lpTimer1->Interval, 0)
END SUB

'-------------------------------------------------------------------------------
' Load components
'-------------------------------------------------------------------------------
SUB LoadComponents
    SendMessage(Application.hMainForm, WM_LOAD_COMPONENTS, 0, 0)
END SUB

'-------------------------------------------------------------------------------
' Remove Timers on Exit
'-------------------------------------------------------------------------------
SUB KillAllTimers
    KillTimer(Application.hMainForm, SYS_TIMER_IDX)
    lpTimer1->Enabled=False
    KillTimer(lpForm1->Handle, lpTimer1->ObjIndex)
END SUB

'-------------------------------------------------------------------------------
' System timer - Display hints
'-------------------------------------------------------------------------------
SUB SysDisplayHint(Parent AS HWND, HintString AS string)
DIM AS TOOLINFO ti

    ti.cbSize=sizeof(ti)
    ti.uFlags=TTF_SUBCLASS: ti.hwnd=Parent
    ti.hinst=NULL: ti.lpszText=STRPTR(HintString)
    GetClientRect(Parent, @ti.rect)

    Application.hWndTTip=CreateWindowEx(WS_EX_TOPMOST, TOOLTIPS_CLASS, "", _
                                    WS_POPUP OR TTS_NOPREFIX OR TTS_ALWAYSTIP, _
                   CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, _
                                                                Parent, 0, 0, 0)
    SetWindowPos(Application.hWndTTip, HWND_TOPMOST, 0, 0, 0, 0, _
                                      SWP_NOMOVE OR SWP_NOSIZE OR SWP_NOACTIVATE)
    SendMessage(Application.hWndTTip, TTM_ADDTOOL, 0, CAST(LPARAM, @ti))	
END SUB

SUB CheckSystemTimers
STATIC AS HWND TT_LastHWND
DIM AS HWND TT_hWnd
DIM AS Point TT_pnt
DIM AS integer ShowIt=False

    TT_pnt.X=Screen.MouseX: TT_pnt.Y=Screen.MouseY
    TT_hWnd=WindowFromPoint(TT_pnt)
    IF TT_LastHWND<>TT_hWnd THEN
        TT_LastHWND=TT_hWnd
        SELECT CASE TT_hWnd
            CASE lpForm1->Handle
                IF lpForm1->ShowHint THEN
                    SysDisplayHint(lpForm1->Handle, lpForm1->Hint)
                    ShowIt=True
                END IF
            CASE lpListBox1->Handle
                IF lpListBox1->ShowHint THEN
                    SysDisplayHint(lpListBox1->Handle, lpListBox1->Hint)
                    ShowIt=True
                END IF
            CASE lpButn2->Handle
                IF lpButn2->ShowHint THEN
                    SysDisplayHint(lpButn2->Handle, lpButn2->Hint)
                    ShowIt=True
                END IF
            CASE lpButn3->Handle
                IF lpButn3->ShowHint THEN
                    SysDisplayHint(lpButn3->Handle, lpButn3->Hint)
                    ShowIt=True
                END IF
        END SELECT
        IF (ShowIt=False) AND (Application.hWndTTip <> 0) THEN
            DestroyWindow(Application.hWndTTip)
            Application.hWndTTip=0
        END IF
    END IF
END SUB

'-------------------------------------------------------------------------------
' Start of Main WndProc code
'-------------------------------------------------------------------------------
SUB CheckWmCommandMsg(hWnd AS HWND, wMsg AS UINT, wParam AS WPARAM, lParam AS LPARAM)
    SELECT CASE LOWORD(wParam)
        CASE lpListBox1->ObjIndex
            IF HIWORD(wParam)=LBN_SELCHANGE THEN
                IF lpListBox1->OnClick>0 THEN lpListBox1->OnClick(lpListBox1)
            END IF
            IF HIWORD(wParam)=LBN_DBLCLK THEN
                IF lpListBox1->OnDblClick>0 THEN lpListBox1->OnDblClick(lpListBox1)
            END IF
        CASE lpButn2->ObjIndex
            IF lpButn2->OnClick>0 THEN lpButn2->OnClick(lpButn2)
        CASE lpButn3->ObjIndex
            IF lpButn3->OnClick>0 THEN lpButn3->OnClick(lpButn3)
    END SELECT
END SUB

SUB CheckWmPaintMsg(hWnd AS HWND, wMsg AS UINT, wParam AS WPARAM, lParam AS LPARAM)
    IF lpForm1->OnPaint>0 THEN lpForm1->OnPaint(lpForm1)
END SUB

SUB CheckWmTimers(hWnd AS HWND, wMsg AS UINT, wParam AS WPARAM, lParam AS LPARAM)
    SELECT CASE wParam
        CASE App_Timer.ObjIndex:  CheckSystemTimers
        CASE lpTimer1->ObjIndex
            IF lpTimer1->OnTimer>0 THEN lpTimer1->OnTimer(lpTimer1)
    END SELECT
END SUB

SUB CheckWmNotifyMsg(hWnd AS HWND, wMsg AS UINT, wParam AS WPARAM, lParam AS LPARAM)
END SUB

SUB CheckWmHitTest(hWnd AS HWND, wMsg AS UINT, wParam AS WPARAM, lParam AS LPARAM)
END SUB

SUB CheckWmObjScroll(hWnd AS HWND, wMsg AS UINT, wParam AS WPARAM, lParam AS LPARAM)
END SUB

SUB CheckWmDrawItem(hWnd AS HWND, wMsg AS UINT, wParam AS WPARAM, lParam AS LPARAM)
END SUB

SUB CheckWmContextMenu(hWnd AS HWND, wMsg AS UINT, wParam AS WPARAM, lParam AS LPARAM)
END SUB

SUB CheckSocketEvent(hWnd AS HWND, wMsg AS UINT, wParam AS WPARAM, lParam AS LPARAM)
' Server: Only ONE message of interest FD_ACCEPT (Server socket ALWAYS non-blocking mode)
' Client: ONLY if YOU set the socket to asynch (non-blocking) the remainder of messages
' NB The wParam is the socket descriptor (handle)
    IF HIWORD(lParam)<>0 THEN
        lpSocket1->SockError=WSAGetLastError()
        EXIT SUB
    END IF
    SELECT CASE lParam
        CASE FD_ACCEPT
            lpSocket1->ConnPending=True                ' Remote connection request
        CASE FD_CONNECT                                '|
        CASE FD_READ                                   '| Client socket asynch mode,...
        CASE FD_WRITE                                  '| ...add your dispatch code here
        CASE FD_CLOSE                                  '|
            lpSocket1->DiscPending=wParam              ' Remote disconnection
    END SELECT
END SUB

FUNCTION WndProc(hWnd AS HWND, wMsg AS UINT, wParam AS WPARAM, lParam AS LPARAM) AS LRESULT
    
    SELECT CASE wMsg
        CASE WM_COMMAND
            IF lParam<>0 THEN
                CheckWmCommandMsg(hWnd, wMsg, wParam, lParam)
            ELSE
                ' IF HIWORD(wParam)=0 THEN
                '     Menu message code goes here.....
                ' ELSE
                '     Accelerator message code goes here...
                ' END IF
            END IF

        CASE WM_PAINT
            CheckWmPaintMsg(hWnd, wMsg, wParam, lParam)

        CASE WM_TIMER
            CheckWmTimers(hWnd, wMsg, wParam, lParam)

        CASE WM_NCHITTEST
            CheckWmHitTest(hWnd, wMsg, wParam, lParam)

        CASE WM_NOTIFY
            CheckWmNotifyMsg(hWnd, wMsg, wParam, lParam)

        CASE WM_HSCROLL, WM_VSCROLL
            IF lParam<>0 THEN CheckWmObjScroll(hWnd, wMsg, wParam, lParam)

        CASE WM_DRAWITEM
            CheckWmDrawItem(hWnd, wMsg, wParam, lParam)

        CASE WM_MOUSEMOVE
            IF lpForm1->OnMouseMove>0 THEN lpForm1->OnMouseMove(LOWORD(lParam), _
                                                            HIWORD(lParam), 0, lpForm1)
        CASE WM_LBUTTONDBLCLK
            IF lpForm1->OnClick>0 THEN lpForm1->OnClick(lpForm1)

        CASE WM_LBUTTONDOWN
            IF lpForm1->OnMouseDown>0 THEN
                lpForm1->OnMouseDown(MK_LBUTTON, LOWORD(lParam), HIWORD(lParam), 0, lpForm1)
            END IF

        CASE WM_LBUTTONUP
            IF lpForm1->OnMouseUp>0 THEN
                lpForm1->OnMouseUp(MK_LBUTTON, LOWORD(lParam), HIWORD(lParam), 0, lpForm1)
            END IF
            IF lpForm1->OnClick>0 THEN lpForm1->OnClick(lpForm1)

        CASE WM_RBUTTONDOWN
            IF lpForm1->OnMouseDown>0 THEN
                lpForm1->OnMouseDown(MK_RBUTTON, LOWORD(lParam), HIWORD(lParam), 0, lpForm1)
            END IF

        CASE WM_RBUTTONUP
            IF lpForm1->OnMouseUp>0 THEN
                lpForm1->OnMouseUp(MK_RBUTTON, LOWORD(lParam), HIWORD(lParam), 0, lpForm1)
            END IF

        CASE WM_CONTEXTMENU
            CheckWmContextMenu(hWnd, wMsg, wParam, lParam)

        CASE WM_KEYDOWN
            IF lpForm1->OnKeyDown>0 THEN lpForm1->OnKeyDown(LOWORD(wParam), 0, lpForm1)

        CASE WM_KEYUP
            IF lpForm1->OnKeyUp>0 THEN lpForm1->OnKeyUp(LOWORD(wParam), 0, lpForm1)

        CASE WM_CHAR
            IF lpForm1->OnKeyPress>0 THEN lpForm1->OnKeyPress(LOBYTE(wParam), lpForm1)

        CASE WM_CREATE

        CASE WM_LOAD_COMPONENTS
            AddChildObjects
            InitSystemTimers(hWnd)

        CASE WM_SOCK_NOTIFY
            CheckSocketEvent(hWnd, wMsg, wParam, lParam)

        CASE WM_CLOSE
            DIM AS integer Action=1                                      ' Default Action is quit
            IF lpForm1->OnClose>0 THEN lpForm1->OnClose(Action, lpForm1) ' Action passed by reference so
            IF Action=0 THEN RETURN 0                                    ' user can change default action

        CASE WM_DESTROY
            KillAllTimers()
            PostQuitMessage(0)

    END SELECT

    RETURN DefWindowProc(hWnd, wMsg, wParam, lParam)
END FUNCTION

'-------------------------------------------------------------------------------
' WndProc hooks for additional components
'-------------------------------------------------------------------------------
'
' ***Add additional WndProcs as required***
'

'-------------------------------------------------------------------------------
' Main startup code
'-------------------------------------------------------------------------------
FUNCTION WinMain(hInst AS HINSTANCE, hPrevInstance AS HINSTANCE, _
                            szCmdLine AS string, iCmdShow AS integer) AS integer
DIM AS MSG wMsg

WITH wc
    .cbSize = SizeOf(WNDCLASSEX)
    .style = CS_HREDRAW OR CS_VREDRAW
    .lpfnWndProc = @WndProc
    .hInstance = hInst
    .hIcon = LoadIcon(0, IDI_APPLICATION)
    .hCursor = LoadCursor(0, IDC_ARROW)
    .hbrBackground = Cast(HBRUSH, COLOR_WINDOW)
    .lpszClassName = @"RQClass"
    .hIconSm = .hIcon
END WITH

    IF RegisterClassEx(@wc) = 0 THEN
        ShowMessage("Register class error")
        RETURN 0
    END IF

    lpForm1->Parent = 0
    lpForm1->Handle = CreateWindowEx(0, "RQClass", lpForm1->Caption, _
                    WS_OVERLAPPEDWINDOW, _
                    lpForm1->Left, lpForm1->Top, lpForm1->Width, lpForm1->Height, _
                    0, 0, hInst, 0)
    Application.hMainForm = lpForm1->Handle

    ' You may wish to delete the following line
    lpForm1->Center

    ' Load remainder of components
    ConfigureComponents()
    LoadComponents()

    IF lpForm1->OnShow>0 THEN lpForm1->OnShow(lpForm1)
    ShowWindow(lpForm1->Handle, SW_SHOW)

    HookTheMouse()
'
'   The following loop is the equivalent of a "Form.ShowModal"
'
    WHILE GetMessage(@wMsg, NULL, 0, 0) <> FALSE
        TranslateMessage(@wMsg)
        DispatchMessage(@wMsg)
    WEND

    UnHookTheMouse()

    RETURN wMsg.wParam
END FUNCTION
'
' Main Entry Point
'
END WinMain(Application.hInstance, NULL, Command(), SW_NORMAL)

'-------------------------------------------------------------------------------
' Code for the component events
'-------------------------------------------------------------------------------
' Called when any client socket disconnects
SUB TidyClientSocketList
DIM AS integer j=0
    FOR i AS integer = 1 TO NumClients
        IF Clients(i)=0 THEN
            ListBox1.AddItems "Removing client #"+STR(i)
            j += 1
        END IF
        IF j>0 THEN Clients(i)=Clients(i+j)
    NEXT i
    IF j>0 THEN NumClients -= j
END SUB

#IF 0
FUNCTION GetPeerIP(Skt AS integer) AS string
DIM arg AS long, res AS long
DIM buf(0 TO 15) AS BYTE
DIM AS string tstr

    arg=16
    res=getpeername(Skt, CAST(sockaddr PTR, @buf(0)), @arg)
    IF res<0 OR arg<1 THEN
        RETURN "<unknown>"
    ELSE
        tstr=""
        FOR i AS integer = 4 TO 7
            res=buf(i)
            IF res<0 THEN res=res AND &Hff
            tstr=tstr & res & "."
        NEXT i
        RETURN LEFT(tstr, LEN(tstr)-1)
    END IF
END FUNCTION
#ENDIF

SUB StartServer(Sender AS QBUTTON)
    Socket1.Parent=Form1.Handle
    ListenSocket=Socket1.sktOpen(SERVER_PORT)    ' Sets the server into listening mode
    IF ListenSocket=0 THEN
        ShowMessage("ERROR "+STR(Socket1.SockError)+": opening server master socket")
        Application.Terminate
    END IF
    ListBox1.AddItems "Server started on "+Socket1.GetHostName()+" ["+Socket1.GetHostIP()+"]"
END SUB

SUB DoTimer(Sender AS QTIMER)
DIM AS integer Sock, Changes=False
DIM AS string buf

    ' First check for any NEW connections
    IF Socket1.ConnectionReady(ListenSocket)>0 THEN
        Sock=Socket1.sktAccept(ListenSocket)
        IF Sock=0 THEN
            ListBox1.AddItems "ERROR "+STR(Socket1.SockError)+": Accept() connection failed"
        ELSE
            buf="Welcome user #"+STR(NumClients+1)
            IF Socket1.WriteLine(Sock, buf)<0 THEN     ' Send welcome message
                ListBox1.AddItems "Client failed to connect!"
                Socket1.sktClose(Sock)
            ELSE
                NumClients += 1
                Clients(NumClients)=Sock
                ListBox1.AddItems "Connected to client #"+STR(NumClients)+" ["+Socket1.ReqPeerName(Sock)+"]"
            END IF
        END IF
    END IF
    ' Now check each connected client for I/C traffic or disconnection
    FOR i AS integer = 1 TO NumClients
        Sock=Clients(i)
        IF Socket1.DiscPending=Sock THEN
            Socket1.DiscPending=0
            ListBox1.AddItems "Client #"+STR(i)+"  disconnected!"
            Socket1.sktClose(Sock)
            Clients(i)=0
            Changes=True
        ELSE
            IF Socket1.IsClientReady(ListenSocket, Sock) THEN        ' I/C traffic?
                buf=Socket1.ReadLine(Sock)   ' Returns ONLY on rcve of a NewLine char or ERROR!!
                IF Socket1.SockError<>0 THEN                         ' Error?
                    ListBox1.AddItems "ERROR "+STR(Socket1.SockError)+": ReadLine() - Client #"+STR(i)
                    Socket1.sktClose(Sock)
                    Clients(i)=0
                    Changes=True
                ELSE
                    buf="From #"+STR(i)+": "+LEFT(buf, LEN(buf)-1)      ' Rcved text message (minus NewLine)
                    ListBox1.AddItems buf
                    FOR j AS integer = 1 TO NumClients                  ' Broadcast the message
                        Sock=Clients(j)
                        IF Sock THEN
                            IF Socket1.WriteLine(Sock, buf)<0 THEN
                                Socket1.sktClose(Sock)
                                Clients(j)=0
                                Changes=True
                            END IF
                        END IF
                    NEXT j
                END IF
            END IF
        END IF
    NEXT i
    IF Changes THEN TidyClientSocketList
END SUB

SUB DoExit(Sender AS QBUTTON)
DIM AS integer Sock
    FOR i AS integer = 1 TO NumClients
        Sock=Clients(i)
        IF Sock > 0 THEN Socket1.sktClose(Sock)
    NEXT i
    Socket1.sktClose(ListenSocket)
    Application.Terminate
END SUB

